#region using
using System;
using System.Collections;
using System.Data.Common;
using System.IO;
using System.Reflection;
using System.Runtime.InteropServices;

using GBPVR.Public;

using DShowNET;
using DShowNET.Device;
using DShowNET.Moonlight;
using DShowNET.GBPVR;
#endregion

namespace SoftwareRecorder
{
	/// <summary>
	/// Summary description for SoftwareRecorder.
	/// </summary>
	public class SoftwareRecorder : IRecordingSource, IRecordingSourceConfiguration
	{

		#region Constructor / Destructor
		public SoftwareRecorder()
		{
			string version = Assembly.GetExecutingAssembly().GetName().Version.ToString();
			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: SoftwareRecorder(version=" + version + ")");
			config = new SoftwareRecorderConfig(this);
		}

		public SoftwareRecorder(SoftwareRecorderConfig config)
		{
			string version = Assembly.GetExecutingAssembly().GetName().Version.ToString();
			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: SoftwareRecorder(version=" + version + ", config=[" + config.GetHashCode() + "])");
			this.config = config;
		}

		~SoftwareRecorder()
		{
			try 
			{
				CloseInterfaces();
			}
			catch {}
			try 
			{
				Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: ~SoftwareRecorder()");
			}
			catch {}
		}
		#endregion

		#region Private Variables

		internal SoftwareRecorderConfig config;

		/// <summary> base filter of the actually used video devices. </summary>
		private IBaseFilter				capVideoFilter;

		/// <summary> base filter of the actually used audio devices. </summary>
		private IBaseFilter				capAudioFilter;

		/// <summary> graph builder interface. </summary>
		private IGraphBuilder			graphBuilder;
		private IntPtr					pGraphBuilder;

		/// <summary> capture graph builder interface. </summary>
		private ICaptureGraphBuilder2	capGraph;

		/// <summary> control interface. </summary>
		private IMediaControl			mediaCtrl;

		/// <summary> crossbars of the capture device. </summary>
		private ArrayList				crossbars = new ArrayList();

		/// <summary> tuner for the capture device. </summary>
		private IAMTVTuner				tuner;

		/// <summary> file writer. </summary>
		private	IFileSinkFilter			sink;

		private	IBaseFilter				videoEncoder;
		private	IBaseFilter				audioEncoder;
		private	IBaseFilter				mux;

		/// <summary> current status of recorder. </summary>
		private	RecordingStatus			currentStatus = RecordingStatus.RECORDING_FAILED;

		private ChannelTuning			currentTuning;
		private int						currentQuality;
		private int						recordingChunkNumber;
		private string					recordingChunkName;
		private string					recordingChunkExtension;
		private long					previousLength;
		private DateTime				chunkLastCheckedAt = DateTime.MinValue;
		private DateTime				statusLastReportedAt = DateTime.MinValue;

#if DEBUG
		/// <summary> the running object table reference. </summary>
		private int						rotCookie = 0;
#endif
		#endregion

		#region Private Methods
		/// <summary> get all the interfaces and setup tuning. </summary>
		void StartupVideo( ChannelTuning channelTuning, bool previewMode, int quality )
		{
			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: enter StartupVideo(channelTuning:={" + channelTuning.analogChannelNumber.ToString() + "," + channelTuning.digitalTuningRequest + "}, previewMode:=" + previewMode + ")");
			try 
			{
				GetCaptureInterfaces();

				CreateCaptureDevices(previewMode, quality);

				if (!previewMode)
				{
					GetRecordInterfaces();

					SetupRecordingGraph();
				}

				setTuning(channelTuning, previewMode);

			}
			catch
			{
				CloseInterfaces();
				throw;
			}
			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: exit StartupVideo");
		}

		/// <summary> create the used COM components and get the interfaces. </summary>
		void GetCaptureInterfaces( )
		{
			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: enter GetCaptureInterfaces()");
			Type comType = null;
			object comObj = null;
			try 
			{
				comType = Type.GetTypeFromCLSID( Clsid.FilterGraph );
				if( comType == null )
					throw new NotImplementedException( @"DirectShow FilterGraph not installed/registered!" );
				comObj = Activator.CreateInstance( comType );
				graphBuilder = (IGraphBuilder) comObj; comObj = null;

#if DEBUG
				DsROT.AddGraphToRot( graphBuilder, out rotCookie );
#endif
			
				comType = Type.GetTypeFromCLSID( Clsid.CaptureGraphBuilder2 );
				if( comType == null )
					throw new NotImplementedException( @"DirectShow CaptureGraphBuilder2 not installed/registered!" );
				comObj = Activator.CreateInstance( comType );
				capGraph = (ICaptureGraphBuilder2) comObj; comObj = null;

				int hr = capGraph.SetFiltergraph( graphBuilder );
				if( hr < 0 ) 
				{
					currentStatus = RecordingStatus.RECORDING_FAILED;
					Marshal.ThrowExceptionForHR( hr );
				}

				mediaCtrl	= (IMediaControl)	graphBuilder;

			}
			finally
			{
				if( comObj != null )
					Marshal.ReleaseComObject( comObj ); comObj = null;
			}
			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: exit GetCaptureInterfaces");
		}

		/// <summary> create the user selected capture device and add it to the graph. </summary>
		void CreateCaptureDevices( bool previewMode, int quality )
		{
			object comObj = null;
			try 
			{
				Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: CreateCaptureDevices-BindToObject(VideoDevice)");

				// create instance of capture device
				comObj = Marshal.BindToMoniker(config.uniqueVideoDeviceIdentifier);
				capVideoFilter = (IBaseFilter) comObj; comObj = null;

				Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: CreateCaptureDevices-IGraphBuilder.AddFilter(VideoFilter)");

				// add video capture device
				int hr = graphBuilder.AddFilter( capVideoFilter, "Video Capture Device" );
				if( hr < 0 ) 
				{
					currentStatus = RecordingStatus.RECORDING_FAILED;
					Marshal.ThrowExceptionForHR( hr );
				}

				IAMAnalogVideoDecoder decoder = capVideoFilter as IAMAnalogVideoDecoder;
				if (decoder != null)
				{
					AnalogVideoStandard oldStandard;
					decoder.get_TVFormat(out oldStandard);
					if (oldStandard != config.videoStandard)
					{
						Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: CreateCaptureDevices-IAMAnalogVideoDecoder.TVFormat=" + config.videoStandard.ToString());
						decoder.put_TVFormat(config.videoStandard);
					}
					decoder = null;
				}

				string profileName = Enum.GetName(typeof(ScheduledRecording_Quality), quality);
				Quality qualityProfile = config.qualityProfiles[profileName] as Quality;
				if (qualityProfile != null)
				{
					Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: CreateCaptureDevices-QualityProfile(" + profileName + ")=" + 
						qualityProfile.colorSpace + "@" + qualityProfile.width + "x" + qualityProfile.height);

					string pinName = previewMode ? config.uniqueVideoDevicePreview : config.uniqueVideoDeviceCapture;
					IPin capturePin;
					hr = DsUtils.FindPin(capVideoFilter, PinDirection.Output, pinName, out capturePin);
					if( hr < 0 ) 
					{
						currentStatus = RecordingStatus.RECORDING_FAILED;
						Marshal.ThrowExceptionForHR( hr );
					}

					DsUtils.DsAMMediaType currentFormat;
					IAMStreamConfig streamConfig = capturePin as IAMStreamConfig;
					DsUtils.GetFormat(streamConfig, out currentFormat);

					Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: CreateCaptureDevices-format before: " + 
						GuidHelper.GetGuidName(typeof(MediaSubType), currentFormat.MediaType.subType) + "@" + currentFormat.videoInfoHeader.BmiHeader.Width + "x" + currentFormat.videoInfoHeader.BmiHeader.Height);

					//DsUtils.ShowCapPinDialog( capGraph, capVideoFilter, IntPtr.Zero );

					currentFormat.MediaType.subType = GuidHelper.GetGuidValue(typeof(MediaSubType), qualityProfile.colorSpace);
					currentFormat.videoInfoHeader.BmiHeader.Width = qualityProfile.width;
					currentFormat.videoInfoHeader.BmiHeader.Height = qualityProfile.height;
					DsUtils.SetFormat(streamConfig, currentFormat);
					
					DsUtils.GetFormat(streamConfig, out currentFormat);

					Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: CreateCaptureDevices-format after: " + 
						GuidHelper.GetGuidName(typeof(MediaSubType), currentFormat.MediaType.subType) + "@" + currentFormat.videoInfoHeader.BmiHeader.Width + "x" + currentFormat.videoInfoHeader.BmiHeader.Height);

				}
				else
				{
					Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: CreateCaptureDevices-QualityProfile(" + profileName + ")=default");
				}
				currentQuality = quality;

				if ((config.uniqueAudioDeviceIdentifier == config.uniqueVideoDeviceIdentifier) || (config.uniqueAudioDeviceIdentifier.Length == 0)) 
				{
					Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: CreateCaptureDevices-AudioInputDevice is VideoInputDevice");
					capAudioFilter = capVideoFilter;
				}
				else
				{
					bool audioPreview = previewMode && !config.audioRendersInPreview;
					if (!audioPreview) 
					{
						Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: CreateCaptureDevices-BindToObject(AudioDevice)");

						// create instance of capture device
						comObj = Marshal.BindToMoniker(config.uniqueAudioDeviceIdentifier);
						capAudioFilter = (IBaseFilter) comObj; comObj = null;

						Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: CreateCaptureDevices-IGraphBuilder.AddFilter(AudioFilter)");
						// add audio capture device
						hr = graphBuilder.AddFilter( capAudioFilter, "Audio Capture Device" );
						if( hr < 0 ) 
						{
							currentStatus = RecordingStatus.RECORDING_FAILED;
							Marshal.ThrowExceptionForHR( hr );
						}

						if ((config.uniqueAudioDeviceSource != null) && (config.uniqueAudioDeviceSource.Length > 0))
						{
							Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: CreateCaptureDevices-GetAudioPins");
							ArrayList pins;
							hr = DsUtils.GetPins(capAudioFilter, PinDirection.Input, out pins);
							if( hr < 0 ) 
							{
								currentStatus = RecordingStatus.RECORDING_FAILED;
								Marshal.ThrowExceptionForHR( hr );
							}

							if (pins != null) 
							{
								Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: CreateCaptureDevices-EnableAudioPin(" + config.uniqueAudioDeviceSource + ")");
								foreach (DsUtils.DsPin pin in pins) 
								{
									bool enable = pin.Name == config.uniqueAudioDeviceSource;
									Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: SetAudioCapturePinState-" + pin.Name + ".Enable=" + enable.ToString());
									(pin.Pin as IAMAudioInputMixer).put_Enable(enable);
									pin.Dispose();
								}
							}
						}

						if (config.videoIsAudioSource)
						{
							Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: CreateCaptureDevices-RenderStream(Video->Audio)");
							// render the video from the capture device via the video encoder to the mux
							Guid cat = Guid.Empty;
							Guid med = Guid.Empty;
							hr = capGraph.RenderStream( ref cat, ref med, capVideoFilter, null, capAudioFilter );
							if( hr < 0 )
							{
								currentStatus = RecordingStatus.RECORDING_FAILED;
								Marshal.ThrowExceptionForHR( hr );
							}
						}
					}

					if ((config.uniqueAudioDeviceSource != null) && (config.uniqueAudioDeviceSource.Length > 0))
					{
						try
						{
							SetAudioPreviewPinState(audioPreview || config.commonRecordingPlaybackMixer);
						}
						catch (Exception ex)
						{
							Logger.Verbose("RecorderConfig[" + this.GetHashCode() + "]: CreateCaptureDevices-failed to SetAudioPreviewPinState: " + ex.ToString());
						}
					}
				}

				if (config.crossbarRoutings.Length > 0) 
				{
					string[] multicrossbarsRoutings = config.crossbarRoutings.Split(';');
					IBaseFilter crossbarSource = capVideoFilter;
					foreach (string multicrossbarRouting in multicrossbarsRoutings)
					{
						Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: CreateCaptureDevices-FindInterface(IAMCrossbar)");

						// find the crossbar - with WDM drivers this should add it automatically to graph for us
						Guid cat  = LookOnly.Upstream;
						Guid type = Guid.Empty;
						Guid iid = typeof( IAMCrossbar ).GUID;
						hr = capGraph.FindInterface( ref cat, ref type, crossbarSource, ref iid, out comObj );
						if( hr < 0 ) 
						{
							currentStatus = RecordingStatus.RECORDING_FAILED;
							Marshal.ThrowExceptionForHR( hr );
						}
						crossbars.Add(comObj);
						IAMCrossbar crossbar = comObj as IAMCrossbar; comObj = null;
						crossbarSource = (IBaseFilter)crossbar;

						Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: CreateCaptureDevices-Crossbar.Route");

						string[] crossbarRoutings = multicrossbarRouting.Split(',');
						foreach (string crossbarRouting in crossbarRoutings)
						{
							string[] routingInfo = crossbarRouting.Split('-');
							int inputPin = Convert.ToInt32(routingInfo[0]);
							int outputPin = Convert.ToInt32(routingInfo[1]);
							int oldInputPin;
							hr = crossbar.get_IsRoutedTo(outputPin, out oldInputPin);
							if( hr < 0 ) 
							{
								currentStatus = RecordingStatus.RECORDING_FAILED;
								Marshal.ThrowExceptionForHR( hr );
							}
							if (oldInputPin != inputPin)
							{
								Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: CreateCaptureDevices-Crossbar.Route(" + inputPin.ToString() + "->" + outputPin.ToString() + ")");
								hr = crossbar.Route(outputPin, inputPin);
								if( hr < 0 ) 
								{
									currentStatus = RecordingStatus.RECORDING_FAILED;
									Marshal.ThrowExceptionForHR( hr );
								}
							}
						}
					}
				}
				else
				{
					Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: CreateCaptureDevices-Not doing Crossbar.Route");
				}
			
				if (config.tuningType != TunerInputType.None) 
				{
					Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: CreateCaptureDevices-FindInterface(IAMTVTuner)");

					// find the tuner - with WDM drivers this should add it automatically to graph for us
					Guid cat  = LookOnly.Upstream;
					Guid type = Guid.Empty;
					Guid iid = typeof( IAMTVTuner ).GUID;
					hr = capGraph.FindInterface( ref cat, ref type, capVideoFilter, ref iid, out comObj );
					if( hr < 0 ) 
					{
						currentStatus = RecordingStatus.RECORDING_FAILED;
						Marshal.ThrowExceptionForHR( hr );
					}
					tuner = comObj as IAMTVTuner; comObj = null;
				}
				else
				{
					Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: CreateCaptureDevices-Not doing FindInterface(IAMTVTuner)");
				}

			}
			finally
			{
				if( comObj != null )
					Marshal.ReleaseComObject( comObj ); comObj = null;
			}
		}

		void SetAudioPreviewPinState(bool enabled)
		{
			string[] deviceIdentifiers = config.uniqueAudioDeviceIdentifier.Split('\\');
			string mixerName = deviceIdentifiers[deviceIdentifiers.Length - 1].Replace(" input", " output");
			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: SetAudioPreviewPinState(mixerName=" + mixerName + ", enabled=" + enabled.ToString());

			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: SetAudioPreviewPinState-WmMixer.GetMixers");

			ArrayList mixers = WmMixer.GetMixers();

			WmMixer mixer = null;
			if (mixers != null) 
			{
				foreach (WmMixer m in mixers)
				{
					Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: SetAudioPreviewPinState-found mixer: " + m.Name);
				}
				foreach (WmMixer m in mixers)
				{
					if (m.Name == mixerName)
					{
						mixer = m;
						break;
					}
				}
			}

			if( mixer == null )
				throw new Exception("Could not find audio mixer device!");

			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: SetAudioPreviewPinState-mixer(" + mixer.Name + ").GetLines");
			ArrayList lines = mixer.GetLines();

			WmLine line = null;
			if (lines != null) 
			{
				foreach (WmLine l in lines)
				{
					Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: SetAudioPreviewPinState-found line: " + l.Name + ", targetName=" + l.Target.Name + ", targetType=" + l.Target.Type.ToString());
				}
				foreach (WmLine l in lines)
				{
					if (l.Target.Type == MixerLine_TargetType.WaveOut) 
					{
						line = l;
						break;
					}
				}
			}

			if( line == null ) 
				throw new Exception("Could not find audio mixer line!");

			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: SetAudioPreviewPinState-line(" + line.Name + ").GetSources");
			ArrayList sources = line.GetSources();

			WmSource source = null;
			if (sources != null) 
			{
				foreach (WmSource s in sources)
				{
					Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: SetAudioPreviewPinState-found source: " + s.Name + ", targetName=" + s.Target.Name + ", targetType=" + s.Target.Type.ToString());
				}
				foreach (WmSource s in sources)
				{
					if (s.Name == config.uniqueAudioDeviceMixer) 
					{
						source = s;
						break;
					}
				}
			}

			if( source == null )
				throw new Exception("Could not find audio mixer source!");

			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: SetAudioPreviewPinState-source(" + source.Name + ").GetLineControls");
			ArrayList lineControls = source.GetLineControls();

			WmLineControl lineControl = null;
			if (lineControls != null) 
			{
				foreach (WmLineControl lc in lineControls)
				{
					Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: SetAudioPreviewPinState-found linecontrol: " + lc.Name + ", type=" + lc.ControlType.ToString());
				}
				foreach (WmLineControl lc in lineControls)
				{
					if (lc.ControlType == MixerControl_ControlType.Mute) 
					{
						lineControl = lc;
						break;
					}
				}
			}

			if( lineControl == null )
				throw new Exception("Could not find audio mixer line control !");

			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: SetAudioPreviewPinState-lineControl(" + lineControl.Name + ").SetValue=" + (!enabled).ToString());

			lineControl.Value = !enabled;

		}

		/// <summary> create the used COM components and get the interfaces. </summary>
		void GetRecordInterfaces( )
		{
			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: enter GetRecordInterfaces()");
			Type comType = null;
			object comObj = null;
			try 
			{
				// create an instace of the GB-PVR file writer sink
				Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: GetRecordInterfaces-CreateInstance(GBPVR.Writer)");
				comType = Type.GetTypeFromCLSID( Clsid_GBPVR.Writer );
				if( comType == null )
					throw new NotImplementedException( @"DirectShow GB-PVR Writer not installed/registered!" );
				comObj = Activator.CreateInstance( comType );
				sink = (IFileSinkFilter) comObj; comObj = null;

				Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: GetRecordInterfaces-BindToObject(VideoEncoder)");
				comObj = Marshal.BindToMoniker(config.uniqueVideoEncoderIdentifier);
				videoEncoder = comObj as IBaseFilter; comObj = null;

				Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: GetRecordInterfaces-BindToObject(AudioEncoder)");
				comObj = Marshal.BindToMoniker(config.uniqueAudioEncoderIdentifier);
				audioEncoder = comObj as IBaseFilter; comObj = null;

				Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: GetRecordInterfaces-BindToObject(Muxer)");
				comObj = Marshal.BindToMoniker(config.uniqueMuxerIdentifier);
				mux = comObj as IBaseFilter; comObj = null;
				
			}
			finally
			{
				if( comObj != null )
					Marshal.ReleaseComObject( comObj ); comObj = null;
			}
			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: exit GetRecordInterfaces");
		}

		/// <summary> build the capture graph for recording. </summary>
		void SetupRecordingGraph()
		{
			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: enter SetupGraph()");

			int hr;
			IPin			muxOutPin		= null;
			object			comObj			= null;

			try 
			{
				Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: SetupGraph-AddFilter(VideoEncoder)");
				// add the video encoder to the graph
				hr = graphBuilder.AddFilter( videoEncoder, "Video Encoder" );
				if( hr < 0 ) 
				{
					currentStatus = RecordingStatus.RECORDING_FAILED;
					Marshal.ThrowExceptionForHR( hr );
				}

				Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: SetupGraph-AddFilter(AudioEncoder)");
				// add the audio encoder to the graph
				hr = graphBuilder.AddFilter( audioEncoder, "Audio Encoder" );
				if( hr < 0 )
				{
					currentStatus = RecordingStatus.RECORDING_FAILED;
					Marshal.ThrowExceptionForHR( hr );
				}

				// the moonlight mux needs it's 2nd stream added manually
				// it also needs to be changed from Transport Stream to Program Stream
				if (mux as IMoonlightConfig != null) 
				{
					Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: SetupGraph-ConfigureMoonlightMux(Muxer)");
					Moonlight.ConfigureMoonlightMux((IMoonlightConfig)mux);
				}

				Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: SetupGraph-AddFilter(Muxer)");
				// add the mux to the graph
				hr = graphBuilder.AddFilter( mux, "Muxer" );
				if( hr < 0 )
				{
					currentStatus = RecordingStatus.RECORDING_FAILED;
					Marshal.ThrowExceptionForHR( hr );
				}

				Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: SetupGraph-AddFilter(Sink)");
				// add the sink to the graph
				hr = graphBuilder.AddFilter( sink as IBaseFilter, null );
				if( hr < 0 )
				{
					currentStatus = RecordingStatus.RECORDING_FAILED;
					Marshal.ThrowExceptionForHR( hr );
				}

				Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: SetupGraph-RenderStream(Video)");
				// render the video from the capture device via the video encoder to the mux
				Guid cat = config.uniqueVideoDeviceCaptureCategory;
				Guid med = MediaType.Video;
				hr = capGraph.RenderStream( ref cat, ref med, capVideoFilter, videoEncoder, mux );
				if( hr < 0 )
				{
					currentStatus = RecordingStatus.RECORDING_FAILED;
					Marshal.ThrowExceptionForHR( hr );
				}

				Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: SetupGraph-RenderStream(Audio)");
				// render the audio from the capture device via the audio encoder to the mux
				cat = config.uniqueAudioDeviceCaptureCategory;
				med = MediaType.Audio;
				hr = capGraph.RenderStream( ref cat, ref med, capAudioFilter, audioEncoder, mux);
				if( hr < 0 )
				{
					currentStatus = RecordingStatus.RECORDING_FAILED;
					Marshal.ThrowExceptionForHR( hr );
				}

				Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: SetupGraph-GetPin(Muxer.Output)");
				// find the output pin of the mux
				hr = DsUtils.GetPin(mux, PinDirection.Output, 0, out muxOutPin);
				if( hr < 0 )
				{
					currentStatus = RecordingStatus.RECORDING_FAILED;
					Marshal.ThrowExceptionForHR( hr );
				}

				Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: SetupGraph-Render(Muxer.Output)");
				// intelligently render the mux output (will connect it to the file sink input pin)
				hr = graphBuilder.Render(muxOutPin);
				if( hr < 0 )
				{
					currentStatus = RecordingStatus.RECORDING_FAILED;
					Marshal.ThrowExceptionForHR( hr );
				}

			}
			finally
			{
				if( muxOutPin != null )
					Marshal.ReleaseComObject( muxOutPin ); muxOutPin = null;
				if( comObj != null ) 
					Marshal.ReleaseComObject( comObj ); comObj = null;
			}
			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: exit SetupGraph");
		}


		/// <summary> do cleanup and release DirectShow. </summary>
		void CloseInterfaces()
		{
			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: enter CloseInterfaces()");
			try 
			{
#if DEBUG 
				if( rotCookie != 0 )
					DsROT.RemoveGraphFromRot( ref rotCookie );
#endif
			}
			catch {}
			try 
			{
				if( mediaCtrl != null )
					mediaCtrl.Stop(); mediaCtrl = null;
			}
			catch {}
			try 
			{
				if( capGraph != null )
					Marshal.ReleaseComObject( capGraph ); capGraph = null;
			}
			catch {}
			try 
			{
				if( pGraphBuilder != IntPtr.Zero )
					Marshal.Release(pGraphBuilder); pGraphBuilder = IntPtr.Zero;
			}
			catch {}
			try 
			{
				if( graphBuilder != null )
					Marshal.ReleaseComObject( graphBuilder ); graphBuilder = null;
			}
			catch {}
			try 
			{
				if( ( capVideoFilter != null ) && ( capVideoFilter != capAudioFilter ) )
					Marshal.ReleaseComObject( capVideoFilter ); capVideoFilter = null;
			}
			catch {}
			try 
			{
				if( capAudioFilter != null )
					Marshal.ReleaseComObject( capAudioFilter ); capAudioFilter = null;
			}
			catch {}
			foreach (object crossbar in (ArrayList)crossbars.Clone())
				try 
				{
					Marshal.ReleaseComObject( crossbar ); crossbars.Remove(crossbar);
				}
				catch {}
			try 
			{
				if( tuner != null )
					Marshal.ReleaseComObject( tuner ); tuner = null;
			}
			catch {}
			try 
			{
				if( sink != null )
					Marshal.ReleaseComObject( sink ); sink = null;
			}
			catch {}
			try 
			{
				if( videoEncoder != null )
					Marshal.ReleaseComObject( videoEncoder ); videoEncoder = null;
			}
			catch {}
			try 
			{
				if( audioEncoder != null ) 
					Marshal.ReleaseComObject( audioEncoder ); audioEncoder = null;
			}
			catch {}
			try 
			{
				if( mux != null )
					Marshal.ReleaseComObject( mux ); mux = null;
			}
			catch {}
			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: exit CloseInterfaces");
		}

		private void setSinkFileName(string fileName)
		{
			// set the filename on the sink
			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: enter setSinkFileName(fileName:=" + fileName + ")");

			previousLength = -1;
			recordingChunkNumber = 0;
			recordingChunkExtension = Path.GetExtension(fileName);
			recordingChunkName = fileName.Substring(0, fileName.Length - recordingChunkExtension.Length);
			while (File.Exists(fileName)) 
			{
				recordingChunkNumber++;
				fileName = recordingChunkName + "-" + recordingChunkNumber + recordingChunkExtension;
			}
			
			AMMediaType mediaType = new AMMediaType();
			int hr = sink.SetFileName(fileName, mediaType);
			if( hr < 0 )
			{
				currentStatus = RecordingStatus.RECORDING_FAILED;
				Marshal.ThrowExceptionForHR( hr );
			}
			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: exit setSinkFileName");
		}

		private void setTuning(ChannelTuning channelTuning, bool previewMode)
		{
			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: enter setTuning(channelTuning:={" + channelTuning.analogChannelNumber.ToString() + "," + channelTuning.digitalTuningRequest + "})");
			if (config.tuningType == TunerInputType.None) 
			{
				Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: setTuning-tuningType is None");
			}
			else
			{
				if (config.fixedChannel != 0) 
				{
					channelTuning.analogChannelNumber = config.fixedChannel;
					Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: setTuning-fixedChannel overides to " + channelTuning.analogChannelNumber.ToString());
				}

				int hr;
				currentTuning = channelTuning;
				AMTunerModeType newMode = AMTunerModeType.TV;

				AMTunerModeType oldMode;
				hr = tuner.get_Mode(out oldMode);
				if (oldMode != newMode) 
				{
					Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: setTuning-IAMTVTuner.Mode=" + oldMode.ToString() + "->" + newMode.ToString());
					hr = tuner.put_Mode(newMode);
					Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: setTuning-put_Mode()=" + hr);
				}

				int oldCountryCode;
				hr = tuner.get_CountryCode(out oldCountryCode);
				if (oldCountryCode != config.countryCode) 
				{
					Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: setTuning-IAMTVTuner.CountryCode=" + oldCountryCode.ToString() + "->" + config.countryCode.ToString());
					hr = tuner.put_CountryCode(config.countryCode);
					Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: setTuning-put_CountryCode()=" + hr);
				}

				TunerInputType oldInput;
				hr = tuner.get_InputType(0, out oldInput);
				if (oldInput != config.tuningType) 
				{
					Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: setTuning-IAMTVTuner.InputType=" + oldInput.ToString() + "->" + config.tuningType.ToString());
					hr = tuner.put_InputType(0, config.tuningType);
					Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: setTuning-put_InputType()=" + hr);
				}
				
				// set the channel
				int oldChannel;
				int oldVideoSubChannel;
				int oldAudioSubChannel;
				hr = tuner.get_Channel(out oldChannel, out oldVideoSubChannel, out oldAudioSubChannel);
				if ((oldChannel != channelTuning.analogChannelNumber) || (!previewMode) || (config.tuneTwice))
				{
					Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: setTuning-IAMTVTuner.Channel=" + oldChannel.ToString() + "->" + channelTuning.analogChannelNumber);
					hr = tuner.put_Channel(channelTuning.analogChannelNumber, AMTunerSubChannel.Default, AMTunerSubChannel.Default);
					Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: setTuning-put_Channel()=" + hr);

					if (config.tuneTwice) 
					{
						hr = tuner.get_Channel(out oldChannel, out oldVideoSubChannel, out oldAudioSubChannel);
						Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: setTuning-IAMTVTuner.Channel=" + oldChannel.ToString() + "->" + channelTuning.analogChannelNumber);
						hr = tuner.put_Channel(channelTuning.analogChannelNumber, AMTunerSubChannel.Default, AMTunerSubChannel.Default);
						Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: setTuning-put_Channel()=" + hr);
					}
				}
			}
			
			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: exit setTuning");
		}
		#endregion

		#region IRecordingSource Members

		public override void startRecording(
			IRecordingSourceConfiguration config, 
			int qualityLevel, 
			string filename, 
			ChannelTuning channelTuning, 
			int minutes, 
			bool useStreamBufferEngineIfAvailable)
		{
			// config == this, pending fix inside of gb-pvr core
			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: enter startRecording(qualityLevel:=" + qualityLevel + ",filename:=" + filename + ",channelTuning:={" + channelTuning.analogChannelNumber.ToString() + "," + channelTuning.digitalTuningRequest + "},minutes:=" + minutes + ",useStreamBufferEngineIfAvailable:=" + useStreamBufferEngineIfAvailable + ")");

			StartupVideo(channelTuning, false, qualityLevel);

			setSinkFileName(filename);

			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: startRecording-IMediaControl.Run");
			int hr = mediaCtrl.Run();
			if( hr < 0 )
			{
				currentStatus = RecordingStatus.RECORDING_FAILED;
				Marshal.ThrowExceptionForHR( hr );
			}
			if (this.config.startRecordTwice)
			{
				hr = mediaCtrl.Stop();
				if( hr < 0 )
				{
					currentStatus = RecordingStatus.RECORDING_FAILED;
					Marshal.ThrowExceptionForHR( hr );
				}
				hr = mediaCtrl.Run();
				if( hr < 0 )
				{
					currentStatus = RecordingStatus.RECORDING_FAILED;
					Marshal.ThrowExceptionForHR( hr );
				}
			}

			currentStatus = RecordingStatus.RECORDING_IN_PROGRESS;

			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: exit startRecording");
		}

		public override void startNewFile(string filename, ChannelTuning channelTuning)
		{
			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: enter startNewFile(filename:=" + filename + ", channelTuning:={" + channelTuning.analogChannelNumber.ToString() + "," + channelTuning.digitalTuningRequest + "})");
			int hr = mediaCtrl.Stop();
			if( hr < 0 )
			{
				currentStatus = RecordingStatus.RECORDING_FAILED;
				Marshal.ThrowExceptionForHR( hr );
			}
			setSinkFileName(filename);
			setTuning(channelTuning, false);
			hr = mediaCtrl.Run();
			if( hr < 0 )
			{
				currentStatus = RecordingStatus.RECORDING_FAILED;
				Marshal.ThrowExceptionForHR( hr );
			}
			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: exit startNewFile");
		}

		public override RecordingStatus getRecordingStatus()
		{
			if (sink != null) 
			{
				if (DateTime.Now > chunkLastCheckedAt.AddSeconds(config.stuckTimeout))
				{
					chunkLastCheckedAt = DateTime.Now;
					string currentFile;
					AMMediaType mediaType;
					sink.GetCurFile(out currentFile, out mediaType);
					FileInfo fileInfo = new FileInfo(currentFile);
					long lengthInMegs = fileInfo.Length / 1048576;

					if (DateTime.Now > statusLastReportedAt.AddSeconds(50))
					{
						// 500 seconds is the EPG update interval, 
						//  the choice of factor is purly cosmetic for the log files
						statusLastReportedAt = DateTime.Now;
						Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: getRecordingStatus, currentStatus=" + currentStatus.ToString() + ", file length=" + lengthInMegs.ToString("#,##0") + " MB");
					}

					if (fileInfo.Length == previousLength) 
					{
						Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: getRecordingStatus-recording stuck");
						int hr;

						Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: getRecordingStatus-IMediaControl.Stop");
						try 
						{
							hr = mediaCtrl.Stop();
							if( hr < 0 )
							{
								currentStatus = RecordingStatus.RECORDING_FAILED;
								Marshal.ThrowExceptionForHR( hr );
							}
						}
						catch (Exception ex)
						{
							Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: getRecordingStatus-catch: " + ex.ToString());
						}
						
						CloseInterfaces();

						if (fileInfo.Length == 0)
						{
							try
							{
								File.Delete(currentFile);
								Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: getRecordingStatus-DeleteEmptyFile");
							}
							catch (Exception e)
							{
								Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: getRecordingStatus-DeleteEmptyFile: " + e.ToString());
							}
						}

						StartupVideo(currentTuning, false, currentQuality);

						setSinkFileName(recordingChunkName + recordingChunkExtension);

						Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: getRecordingStatus-IMediaControl.Run");
						hr = mediaCtrl.Run();
						if( hr < 0 )
						{
							currentStatus = RecordingStatus.RECORDING_FAILED;
							Marshal.ThrowExceptionForHR( hr );
						}
					} 
					else
					{
						previousLength = fileInfo.Length;
					}
				}
			}
			return currentStatus;
		}

		public void Poke()
		{
			int hr = mediaCtrl.Stop();
			if( hr < 0 )
			{
				Marshal.ThrowExceptionForHR( hr );
			}
		}

		public override void stopRecording()
		{
			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: enter stopRecording()");
			if ((config.uniqueAudioDeviceIdentifier != config.uniqueVideoDeviceIdentifier) && (config.uniqueAudioDeviceSource != null) && (config.uniqueAudioDeviceSource.Length > 0))
				try 
				{
					SetAudioPreviewPinState(false);
				}
				catch (Exception ex)
				{
					Logger.Verbose("RecorderConfig[" + this.GetHashCode() + "]: stopRecording-failed to SetAudioPreviewPinState: " + ex.ToString());
				}
			if (mediaCtrl != null) 
			{
				try 
				{
					int hr = mediaCtrl.Stop();
					if( hr < 0 )
					{
						Marshal.ThrowExceptionForHR( hr );
					}
				}
				catch (Exception ex)
				{
					Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: stopRecording-catch: " + ex.ToString());
				}
				currentStatus = RecordingStatus.RECORDING_COMPETED;
			}
			CloseInterfaces();
			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: exit stopRecording");
		}

		public override string getFinalFileName()
		{
			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: enter getFinalFileName()");
			string currentFile;
			AMMediaType mediaType;
			sink.GetCurFile(out currentFile, out mediaType);
			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: exit getFinalFileName, return " + currentFile);
			return currentFile;
		}

		public override bool canRecordFMRadio()
		{
			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: canRecordFMRadio, return false");
			return false;
		}

		public override void startFMRecording(string filename, int freq)
		{
			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: startFMRecording, throw NotImplementedException");
			throw new NotImplementedException("Currently not implemented");
		}

		public override string getUniqueDeviceIdentifier()
		{
			//Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: getUniqueDeviceIdentifier, return " + this.GetType().Name + ":" + config.captureSourceOID);
			return this.GetType().Name + ":" + config.captureSourceOID;
		}

		public override PreviewGraph getPreviewGraph(ChannelTuning channelTuning)
		{
			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: enter getPreviewGraph(channelTuning:={" + channelTuning.analogChannelNumber.ToString() + "," + channelTuning.digitalTuningRequest + "})");
			
			StartupVideo(channelTuning, true, (int)ScheduledRecording_Quality.LiveTVPreview);

			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: getPreviewGraph-GetComInterfaceForObject:IGraphBuilder");
			pGraphBuilder = Marshal.GetComInterfaceForObject(graphBuilder, typeof(IGraphBuilder));

			FilterInfo videoInfo = new FilterInfo();
			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: getPreviewGraph-VideoFilter:QueryFilterInfo");
			int hr = capVideoFilter.QueryFilterInfo(videoInfo);
			if( hr < 0 ) 
			{
				currentStatus = RecordingStatus.RECORDING_FAILED;
				Marshal.ThrowExceptionForHR( hr );
			}
			
			FilterInfo audioInfo = new FilterInfo();
			if (capAudioFilter != null) 
			{
				Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: getPreviewGraph-AudioFilter:QueryFilterInfo");
				hr = capAudioFilter.QueryFilterInfo(audioInfo);
				if( hr < 0 ) 
				{
					currentStatus = RecordingStatus.RECORDING_FAILED;
					Marshal.ThrowExceptionForHR( hr );
				}
			}

			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: getPreviewGraph-new PreviewGraph()");
			PreviewGraph graph = new PreviewGraph();
			graph.graphBuilder = pGraphBuilder;
			graph.sourceCaptureVideoFilterName = videoInfo.achName;
			graph.sourceCaptureVideoPinName = config.uniqueVideoDeviceCapture;
			graph.sourcePreviewVideoFilterName = videoInfo.achName;
			graph.sourcePreviewVideoPinName = config.uniqueVideoDevicePreview;
			graph.sourceAudioFilterName = audioInfo.achName;
			graph.sourceAudioPinName = config.uniqueAudioDeviceCapture;

			if( videoInfo.pGraph != null )
				Marshal.ReleaseComObject(videoInfo.pGraph);
			if( audioInfo.pGraph != null )
				Marshal.ReleaseComObject(audioInfo.pGraph);

			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: exit getPreviewGraph, return [" + graph.GetHashCode() + "]");
			return graph;
		}

		public override void setPreviewToChannel(ChannelTuning channelTuning)
		{
			Logger.Verbose("SoftwareRecorder[" + this.GetHashCode() + "]: setPreviewToChannel(channelTuning:={" + channelTuning.analogChannelNumber.ToString() + "," + channelTuning.digitalTuningRequest + "})");

			int hr;
			switch (config.channelChangeOperation)
			{
				case ChannelChangeOperation.Pause:
					hr = mediaCtrl.Pause();
					break;
				case ChannelChangeOperation.Stop:
					hr = mediaCtrl.Stop();
					break;
				case ChannelChangeOperation.StopWhenReady:
					hr = mediaCtrl.StopWhenReady();
					break;
				default:
					hr = 0;
					break;
			}
			if( hr < 0 ) 
			{
				currentStatus = RecordingStatus.RECORDING_FAILED;
				Marshal.ThrowExceptionForHR( hr );
			}
			
			setTuning(channelTuning, true);
			
			switch (config.channelChangeOperation)
			{
				case ChannelChangeOperation.Pause:
					hr = mediaCtrl.Run();
					break;
				case ChannelChangeOperation.Stop:
					hr = mediaCtrl.Run();
					break;
				case ChannelChangeOperation.StopWhenReady:
					hr = mediaCtrl.Run();
					break;
				default:
					hr = 0;
					break;
			}
			if( hr < 0 ) 
			{
				currentStatus = RecordingStatus.RECORDING_FAILED;
				Marshal.ThrowExceptionForHR( hr );
			}
		}

		#endregion

		#region IRecordingSourceConfiguration Pass-Through Members

		public void Save(DbConnection connection, int captureSourceOID)
		{
			config.Save(connection, captureSourceOID);
		}

		public void Delete(DbConnection connection, int captureSourceOID)
		{
			config.Delete(connection, captureSourceOID);
		}

		public void Load(int captureSourceOID)
		{
			config.Load(captureSourceOID);
		}

		public string getName()
		{
			return config.getName() + " (broken plugin loading pass-through)";
		}

		public string getConfigurationFormClassName()
		{
			return config.getConfigurationFormClassName();
		}

		public IRecordingSource getRecordingSource()
		{
			return config.getRecordingSource();
		}

		#endregion

	}

}
